/***************************************************************
 *  ReVOLUX Version 0.1                                        *
 *  Music Extractor for the PSX Dance Dance Revolution Series  *
 *                                                             *
 *  Creator: FadinhaX (cantremember@zipmail.com.br)            *
 *  Date: August 1st, 2002                                     *
 *                                                             *
 *  Based on the VAG decoding code by bITmASTER@bigfoot.com    *
 ***************************************************************/




#include <math.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>




#define WavHeader "RIFF????WAVEfmt \020\000\000\000\001\000\002\000\104\254\000\000\020\261\002\000\004\000\020\000data????"




const double Predict[5][2] = { {          0.0,          0.0 },
                               {  60.0 / 64.0,          0.0 },
                               { 115.0 / 64.0, -52.0 / 64.0 },
                               {  98.0 / 64.0, -55.0 / 64.0 },
                               { 122.0 / 64.0, -60.0 / 64.0 } };




int  ScanVag( FILE *VagFileHandle );
long DecodeVag( char *VagBuffer, short *WavBuffer, double *Sample1, double *Sample2 );




int main( int argc, char **argv )
{
 FILE  *VagFileHandle;
 FILE  *WavFileHandle;

 char   VagBuffer[0x8000];
 short  WavBuffer[0x1C000];

 short  TotalChunksDecoded;
 short  WavFileNumber;
 long   WavFileSize;
 char   WavFileName[11];

 double LSample1;
 double LSample2;
 double RSample1;
 double RSample2;



 printf( "ReVOLUX Version 0.1\n" );
 printf( "Music Extractor for the PSX Dance Dance Revolution Series!\n" );
 printf( "Created by FadinhaX (cantremember@zipmail.com.br) on August 1st, 2002.\n\n" );

 if( argc != 2 ) {
    printf( "Usage: ReVOLUX [CD-ROM:]\\STR.BIN\n" );
    return 1;
  }

 WavFileNumber = 0;
 VagFileHandle = fopen( argv[1], "rb" );

 while( ScanVag(VagFileHandle) ) {
    sprintf( WavFileName, "MUSIC-%02d.WAV", WavFileNumber );
    printf( "Extracting %s ... ", WavFileName );
    WavFileNumber++;

    WavFileHandle = fopen( WavFileName, "wb" );
    fwrite( WavHeader, 44, 1, WavFileHandle );

    LSample1 = 0;
    LSample2 = 0;
    RSample1 = 0;
    RSample2 = 0;

    do {
       fread( VagBuffer, 0x8000, 1, VagFileHandle );
       TotalChunksDecoded  = DecodeVag( VagBuffer, WavBuffer, &LSample1, &LSample2 );               /*  Left Channel */
       TotalChunksDecoded += DecodeVag( VagBuffer + 0x4000, WavBuffer + 1, &RSample1, &RSample2 );  /* Right Channel */
       fwrite( WavBuffer, TotalChunksDecoded * 28, 2, WavFileHandle );
     } while( TotalChunksDecoded == 0x800 );

    fseek( WavFileHandle, 0, SEEK_END );
    WavFileSize = ftell( WavFileHandle );

    WavFileSize -= 8;
    fseek( WavFileHandle, 4, SEEK_SET );
    fwrite( &WavFileSize, 4, 1, WavFileHandle );

    WavFileSize -= 36;
    fseek( WavFileHandle, 32, SEEK_CUR );
    fwrite( &WavFileSize, 4, 1, WavFileHandle );

    fclose( WavFileHandle );
    printf( "ok!\n" );
  }

 fclose( VagFileHandle );

 printf( "\nFinished! %02d music(s) extracted.\n", WavFileNumber );
 return 0;
}




int ScanVag( FILE *VagFileHandle )
{
 char SearchBuffer[0x10];



 while( !feof(VagFileHandle) ) {
    fread( SearchBuffer, 0x10, 1, VagFileHandle );
    if( SearchBuffer[1] == 0x06 ) {
       fseek( VagFileHandle, -0x20, SEEK_CUR );
       return 1;
     }
  }

 return 0;
}




long DecodeVag( char *VagBuffer, short *WavBuffer, double *Sample1, double *Sample2 )
{
 short   Chunks;
 short   Samples;
 long    ActualSample;
 long    PredictFactor;
 long    ShiftFactor;

 double  SamplesVector[28];
 double *SamplesPointer;



 for( Chunks = 0; Chunks < 0x400; Chunks++ ) {
    PredictFactor = *VagBuffer >> 4;
    ShiftFactor   = *VagBuffer  & 0x0F;

    VagBuffer++;
    if( *VagBuffer == 0x01 )
       break;

    VagBuffer++;
    SamplesPointer = SamplesVector;
    for( Samples = 0; Samples < 14; Samples++ ) {

       ActualSample = ( *VagBuffer & 0x0F ) << 12;
       if( ActualSample & 0x8000 )
           ActualSample |= 0xFFFF0000;
      *SamplesPointer = (double)( ActualSample >> ShiftFactor );
       SamplesPointer++;

       ActualSample = ( *VagBuffer & 0xF0 ) << 8;
       if( ActualSample & 0x8000 )
           ActualSample |= 0xFFFF0000;
      *SamplesPointer = (double)( ActualSample >> ShiftFactor );
       SamplesPointer++;

       VagBuffer++;
     }

    SamplesPointer = SamplesVector;
    for( Samples = 0; Samples < 28; Samples++ ) {
      *SamplesPointer += *Sample1 * Predict[PredictFactor][0] + *Sample2 * Predict[PredictFactor][1];
      *Sample2 = *Sample1;
      *Sample1 = *SamplesPointer;

      *SamplesPointer = ceil( *SamplesPointer );
       if( *SamplesPointer >  32767 )  *SamplesPointer =  32767;
       if( *SamplesPointer < -32768 )  *SamplesPointer = -32768;

      *WavBuffer  = (short)( *SamplesPointer );
       WavBuffer += 2;
       SamplesPointer++;
     }
  }     

 return Chunks;
}
