/***************************************************************
 *  ReVOLUX Version 0.2                                        *
 *  Music Extractor for the PSX Dance Dance Revolution Series  *
 *                                                             *
 *  Creator: FadinhaX (cantremember@zipmail.com.br)            *
 *  Date: August 3rd, 2002                                     *
 *                                                             *
 *  Based on the VAG decoding code by bITmASTER@bigfoot.com    *
 ***************************************************************/




#include <math.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>



#define  VAG_CHUNKDATA        0xE
#define  VAG_CHUNKSIZE        0x10
#define  VAG_BLOCKSIZE        0x8000
#define  VAG_CHUNKSPERBLOCK   0x800

#define  DECODE_CHUNKSIZE     0x38
#define  DECODE_BLOCKSIZE     0x1C000


#define  VAG_MEMORYBUFFER     0x100000
#define  VAG_MAXBLOCKS        VAG_MEMORYBUFFER / VAG_BLOCKSIZE
#define  VAG_MAXCHUNKS        VAG_MAXBLOCKS * VAG_CHUNKSPERBLOCK

#define  DECODE_MEMORYBUFFER  VAG_MAXBLOCKS * DECODE_BLOCKSIZE


#define  WAV_HEADER           "RIFF????WAVEfmt \020\000\000\000\001\000\002\000\104\254\000\000\020\261\002\000\004\000\020\000data????"
#define  WAV_HEADERSIZE       0x2C

#define  ERROR_OPEN           0x01
#define  ERROR_READ           0x02
#define  ERROR_WRITE          0x04
#define  ERROR_CREATE         0x08
#define  ERROR_HEADER         0x10
#define  ERROR_MEMORY         0x20
#define  ERROR_HEADERUPDATE   0x40

#define  TRUE                 1
#define  FALSE                0
#define  FOUND                TRUE
#define  NOTFOUND             FALSE




void  CheckArguments            ( int argc, char **argv );
void  ClearAllocatedMemory      ( void );
void  CloseOpenFileHandles      ( void );
void  CreateWavFile             ( void );
void  DecodeHalfVagBlock        ( double *Sample1, double *Sample2 );
void  DecodeVagMusic            ( void );
void  Initialize                ( void );
void  InitializeDecodeFunction  ( void );
void  Error                     ( int ErrorType, char *ErrorFileName );
long  FillVagBuffer             ( void );
void  Finalize                  ( void );
void  FinalizeDecodeFunction    ( void );
int   SearchVagMusic            ( void );
void  UpdateWavFileHeader       ( void );
void  WelcomeMessage            ( void );
void  WriteWavFile              ( void );




FILE   *VagFile;
FILE   *WavFile;

char   *VagFileName;
char    WavFileName[11];
int     WavFileNumber;

char   *VagBuffer;
char   *VagPointer;
short  *DecodeBuffer;
short  *DecodePointer;

int     LastBlock;
int     DecodedBlocks;
long    DecodedChunks;

long    BytesRead;
long    BytesWritten;

double  LChannelSample1;
double  LChannelSample2;
double  RChannelSample1;
double  RChannelSample2;

const double  Predict[5][2] = { {          0.0,          0.0 },
                                {  60.0 / 64.0,          0.0 },
                                { 115.0 / 64.0, -52.0 / 64.0 },
                                {  98.0 / 64.0, -55.0 / 64.0 },
                                { 122.0 / 64.0, -60.0 / 64.0 } };




int main( int argc, char **argv )
{
 WelcomeMessage( );

 CheckArguments( argc, argv );
 Initialize( );

 while( SearchVagMusic() == FOUND ) {
    CreateWavFile( );
    InitializeDecodeFunction( );

    do {
       FillVagBuffer( );
       DecodeVagMusic( );
       WriteWavFile( );
     } while( DecodedChunks == VAG_MAXCHUNKS );

    FinalizeDecodeFunction( );
    UpdateWavFileHeader( );
  }

 Finalize( );
 return 0;
}




void CheckArguments( int argc, char **argv )
{
 if( argc != 2 ) {
    printf( "Usage: ReVOLUX [CD-ROM:]\\STR.BIN\n" );
    exit( 1 );
  }

 VagFileName = argv[1];
 return;
}




void ClearAllocatedMemory( void )
{
 if( VagBuffer != NULL )
    free( VagBuffer );

 if( DecodeBuffer != NULL )
    free( DecodeBuffer );

 return;
}




void CloseOpenFileHandles( void )
{
 if( VagFile != NULL )
    fclose( VagFile );

 if( WavFile != NULL )
    fclose( WavFile );

 return;
}




void CreateWavFile( void )
{
 sprintf( WavFileName, "MUSIC-%02d.WAV", WavFileNumber );
 printf( "Extracting %s ... ", WavFileName );
 WavFileNumber++;

 WavFile = fopen( WavFileName, "wb" );
 if( WavFile == NULL )
    Error( ERROR_CREATE, WavFileName );

 BytesWritten = fwrite( WAV_HEADER, 1, WAV_HEADERSIZE, WavFile );
 if( BytesWritten != WAV_HEADERSIZE )
    Error( ERROR_HEADER, WavFileName );

 return;
}




void DecodeHalfVagBlock( double *Sample1, double *Sample2 )
{
 int     Chunks;
 int     Samples;
 int     ShiftFactor;
 int     PredictFactor;
 long    CurrentSample;

 double  SamplesVector[28];
 double *SamplesPointer;


 for( Chunks = 0; Chunks < VAG_CHUNKSPERBLOCK / 2; Chunks++ ) {
    PredictFactor = *VagPointer >> 0x04;
    ShiftFactor   = *VagPointer  & 0x0F;
    VagPointer++;

    if( *VagPointer == 0x01 ) {
       LastBlock = TRUE;
       break;
     }

    VagPointer++;
    SamplesPointer = SamplesVector;
    for( Samples = 0; Samples < VAG_CHUNKDATA; Samples++ ) {
       CurrentSample = ( *VagPointer & 0x0F ) << 0x0C;
       if( CurrentSample & 0x8000 )
          CurrentSample |= 0xFFFF0000;
      *SamplesPointer = (double)( CurrentSample >> ShiftFactor );
       SamplesPointer++;

       CurrentSample = ( *VagPointer & 0xF0 ) << 0x08;
       if( CurrentSample & 0x8000 )
          CurrentSample |= 0xFFFF0000;
      *SamplesPointer = (double)( CurrentSample >> ShiftFactor );
       SamplesPointer++;

       VagPointer++;
     }

    SamplesPointer = SamplesVector;
    for( Samples = 0; Samples < VAG_CHUNKDATA * 2; Samples++ ) {
      *SamplesPointer += *Sample1 * Predict[PredictFactor][0] + *Sample2 * Predict[PredictFactor][1];
      *Sample2 = *Sample1;
      *Sample1 = *SamplesPointer;

      *SamplesPointer = ceil( *SamplesPointer );
       if( *SamplesPointer >  32767 )
           *SamplesPointer =  32767;
       if( *SamplesPointer < -32768 )
           *SamplesPointer = -32768;

      *DecodePointer  = (short) *SamplesPointer;
       DecodePointer += 0x02;
       SamplesPointer++;
     }
  }

 DecodedChunks += Chunks;
 return;
}




void DecodeVagMusic( void )
{
 DecodedChunks = 0;

 for( DecodedBlocks = 0; DecodedBlocks < VAG_MAXBLOCKS; DecodedBlocks++ ) {
    VagPointer    = VagBuffer + DecodedBlocks * VAG_BLOCKSIZE;
    DecodePointer = DecodeBuffer + DecodedBlocks * DECODE_BLOCKSIZE / 2;
    DecodeHalfVagBlock( &LChannelSample1, &LChannelSample2 );

    VagPointer    = VagBuffer + DecodedBlocks * VAG_BLOCKSIZE + VAG_BLOCKSIZE / 2;
    DecodePointer = DecodeBuffer + DecodedBlocks * DECODE_BLOCKSIZE / 2 + 0x01;
    DecodeHalfVagBlock( &RChannelSample1, &RChannelSample2 );

    if( LastBlock == TRUE ) {
       DecodedBlocks++;
       break;
     }
  }

 return;
}




void Error( int ErrorType, char *ErrorFileName )
{
 char *ErrorMessage;


 switch( ErrorType ) {
    case ERROR_OPEN:         ErrorMessage = "Error: Cannot open file %s\n";                  break;
    case ERROR_READ:         ErrorMessage = "Error: Cannot read music data from file %s\n";  break;
    case ERROR_WRITE:        ErrorMessage = "Error: Cannot write music data to file %s\n";   break;
    case ERROR_CREATE:       ErrorMessage = "Error: Cannot create file %s\n";                break;
    case ERROR_HEADER:       ErrorMessage = "Error: Cannot write wav header to file %s\n";   break;
    case ERROR_MEMORY:       ErrorMessage = "Error: Cannot allocate necessary memory\n";     break;
    case ERROR_HEADERUPDATE: ErrorMessage = "Error: Cannot update wav header on file %s\n";  break;
  }

 printf( ErrorMessage, ErrorFileName );
 ClearAllocatedMemory( );
 CloseOpenFileHandles( );

 if( ErrorType & (ERROR_WRITE | ERROR_HEADER | ERROR_HEADERUPDATE) )
    unlink( ErrorFileName );

 exit( 1 );
}




long FillVagBuffer( void )
{
 BytesRead = fread( VagBuffer, 1, VAG_MEMORYBUFFER, VagFile );
 return BytesRead;
}




void Finalize( void )
{
 printf( "\nFinished! %02d music(s) extracted.\n", WavFileNumber );

 ClearAllocatedMemory( );
 CloseOpenFileHandles( );
 return;
}




void FinalizeDecodeFunction( void )
{
 fseek( VagFile, (DecodedBlocks - VAG_MAXBLOCKS) * VAG_BLOCKSIZE, SEEK_CUR );
 return;
}




void Initialize( void )
{
 VagFile = fopen( VagFileName, "rb" );
 if( VagFile == NULL )
    Error( ERROR_OPEN, VagFileName );

 VagBuffer = malloc( VAG_MEMORYBUFFER );
 if( VagBuffer == NULL )
    Error( ERROR_MEMORY, NULL );

 DecodeBuffer = malloc( DECODE_MEMORYBUFFER );
 if( DecodeBuffer == NULL )
    Error( ERROR_MEMORY, NULL );

 WavFileNumber = 0;
 return;
}




void InitializeDecodeFunction( void )
{
 LChannelSample1 = 0;
 LChannelSample1 = 0;
 RChannelSample1 = 0;
 RChannelSample1 = 0;
 LastBlock       = FALSE;

 return;
}




int SearchVagMusic( void )
{
 while( BytesRead = FillVagBuffer() ) {
    VagPointer = VagBuffer + 0x01;

    do {
       if( *VagPointer == 0x06 ) {
          fseek( VagFile, - BytesRead - VAG_CHUNKSIZE, SEEK_CUR );
          return FOUND;
        }

       VagPointer += VAG_CHUNKSIZE;
       BytesRead  -= VAG_CHUNKSIZE;
     } while( BytesRead > 0 );
  }

 return NOTFOUND;
}




void UpdateWavFileHeader( void )
{
 long  WavFileSize;


 fseek( WavFile, 0, SEEK_END );
 WavFileSize = ftell( WavFile );

 WavFileSize -= 0x08;
 fseek( WavFile, 0x04, SEEK_SET );
 BytesWritten = fwrite( &WavFileSize, 1, 0x04, WavFile );
 if( BytesWritten != 0x04 )
    Error( ERROR_HEADERUPDATE, WavFileName );

 WavFileSize -= 0x24;
 fseek( WavFile, 0x20, SEEK_CUR );
 BytesWritten = fwrite( &WavFileSize, 1, 0x04, WavFile );
 if( BytesWritten != 0x04 )
    Error( ERROR_HEADERUPDATE, WavFileName );

 fclose( WavFile );
 printf( "ok!\n" );

 return;
}




void WelcomeMessage( void )
{
 printf( "ReVOLUX Version 0.2\n" );
 printf( "Music Extractor for the PSX Dance Dance Revolution Series!\n" );
 printf( "Created by FadinhaX (cantremember@zipmail.com.br) on August 3rd, 2002.\n\n" );

 return;
}




void WriteWavFile( void )
{
 long  BytesToWrite;


 BytesToWrite = DecodedChunks * DECODE_CHUNKSIZE;

 BytesWritten = fwrite( DecodeBuffer, 1, BytesToWrite, WavFile );
 if( BytesWritten != BytesToWrite )
    Error( ERROR_WRITE, WavFileName );

 return;
}